//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KeyingRhythm class - records keystroke timings, and calculates
// statistical values when requested
//

/**
 * Records keystroke timings, and calculates statistical values 
 * when requested.
 */
public class KeyingRhythm {

	/**
	 * Default size of a timing circular buffer.
	 */
	public static final int	DEFAULT_LENGTH		= 1000;

	/**
	 * Default tolerance percentage on calculating rhythm measure.
	 */
	public static final int	DEFAULT_TOLERANCE	= 20;

	// array of timings, used as a circular buffer
	private int		buffer[];//		= null;

	// index into the timings array, for circular buffer-ness, and
	// note if the index wrapped round at any time
	private int		buffer_index;//		= 0;
	private int		buffer_wraps;//		= 0;


	/**
	 * Create a new rhythm monitor, with the specified buffer length.
	 */
	public KeyingRhythm( int length ) {

		// create the new array
		buffer		= new int[ length ];
		buffer_index	= 0; buffer_wraps	= 0;
	}

	/**
	 * Create a new rhythm monitor, with the default buffer length
	 * of 1000.
	 */
	public KeyingRhythm() {
		this( DEFAULT_LENGTH );
	}

	/**
	 * Record a new keystroke timing.
	 */
	public void struck( int msec ) {

		// add the time to the buffer, and increment the index
		buffer[ buffer_index++ ] = msec;

		// wrap the index if it is now beyond the array
		if ( buffer_index >= buffer.length ) {
			buffer_index	= 0; buffer_wraps++;
		}
	}

	/**
	 * Reset the monitor to the initial state.
	 */
	public void reset() {
		buffer_index	= 0; buffer_wraps	= 0;
	}

	/**
	 * Calculate a % rhythm accuracy from the information we have recorded,
	 * by counting the number of timings that fall within +/-tolerance%
	 * of the average timing.
	 */
	public int get_pc_rhythm( int tolerance ) {

		// see if any values were recorded at all
		if ( buffer_index > 0 || buffer_wraps > 0 ) {

			// normalize the tolerance parameter
			if ( tolerance > 100 )
				tolerance = 100;
			else if ( tolerance < 0 )
				tolerance = 0;

			// determine how much of the buffer to use
			int buffer_max	= ( buffer_wraps > 0 ) ?
						buffer.length : buffer_index;

			// find the median value for this data set; start by
			// find the largest and smallest values in it
			int maximum	= Integer.MIN_VALUE;
			int minimum	= Integer.MAX_VALUE;
			for ( int index = 0; index < buffer_max; index++ ) {
				maximum = Math.max( maximum, buffer[ index ] );
				minimum = Math.min( minimum, buffer[ index ] );
			}

			// now use a convergent loop to get to the median;
			// start with an estimated median at the middle of the
			// range, and a delta that is half of the range, then
			// loop until the delta is less than one - a fine point
			// to stop since we are dealing with integers
			float median	= (float)( maximum + minimum ) / 2.0f;
			float delta	= (float)( maximum - minimum ) / 2.0f;
			while ( delta >= 1.0f ) {

				// count the number of points below the current
				// estimate of the median
				int less_count = 0;
				for ( int index = 0; index < buffer_max;
						index++ ) {
					if ( (float)(buffer[ index ]) < median )
						less_count++;
				}

				// adjust the median estimate, depending on
				// how the count of points came out, or stop
				// if we've found the median
				if ( less_count > ( buffer_max / 2 ) )
					median -= delta;
				else if ( less_count < ( buffer_max / 2 ) )
					median += delta;
				else
					break;

				// halve the delta for the next loop
				delta = delta / 2.0f;
			}

			// convert the median-ish value found above into
			// an integer; since all the data values are integral,
			// the median should be too - the above loop may have
			// found a median, but it could be +/- 0.5 inaccurate
			median = (float)( (int)( median + 0.5f ) );

			// now, count how many data points fall inside the
			// given tolerance band of that median
			float deviation	= median * (float)( tolerance ) / 100f;
			float band_min	= median - deviation;
			float band_max	= median + deviation;
			int inside_count= 0;
			for ( int index = 0; index < buffer_max; index++ ) {
				float value	= (float)buffer[ index ];
				if ( value >= band_min && value <= band_max )
					inside_count++;
			}

			// return a measure of the rhythm regularity
			return ( ( inside_count * 100 ) / buffer_max );
		}
		else
			// no timings measured
			return ( 0 );
	}

	/**
	 * Calculate a % rhythm using a default tolerance of +/-10%.
	 */
	public int get_pc_rhythm() {
		return ( get_pc_rhythm( DEFAULT_TOLERANCE ));
	}


	/**
	 * Printout method.
	 */
	public String toString() {
		return ( getClass().getName() + "{buffer_index=" + buffer_index
				+ ",buffer_wraps=" + buffer_wraps
				+ ",get_pc_rhythm()=" + get_pc_rhythm() + "}" );
	}


	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		KeyingRhythm g = new KeyingRhythm();
		System.out.println( "initial : " + g );
		g.struck( 200 ); System.out.println( "200mS : " + g );
		g.struck( 100 ); g.struck( 300 );
		System.out.println( "1/2/300mS : " + g );
		g.reset(); System.out.println( "reset : " + g );
		for ( int i = 0; i < 100; i++ ) {
			g.struck( 100 ); g.struck( 300 );
		}
		g.reset(); for ( int i = 0; i < 100; i++ ) {
			g.struck( 195 ); g.struck( 205 );
			g.struck( 190 ); g.struck( 210 );
			g.struck( 180 ); g.struck( 220 );
		}
		System.out.println( "reasonable : " + g );
		g.reset(); for ( int i = 100; i < 1334; i++ )
			g.struck( i );
		System.out.println( "full : " + g );
	}
}
